/*!***************************************************************************
*!
*! FILE NAME  : cgiparse.c (based on http_pkt.c)
*!
*! DESCRIPTION: Parse HTTP CGI queries and headers.
*!
*! FUNCTIONS  : url_parameter_type_init
*! (EXTERNAL)   url_parameter_type_init_header
*!              url_parameter_type_done
*!              url_parameter_get_data
*!
*! FUNCTIONS  : url_param_fix
*! (LOCAL)      
*!
*! ---------------------------------------------------------------------------
*! HISTORY
*!
*! DATE         NAME               CHANGES
*! ----         ----               -------
*! Feb 12 2000  Johan Adolfsson    Initial version
*!
*! This file is distributed under the Gnu Public License (GPL),
*! please see the file LICENSE for further information.
*!
*! ---------------------------------------------------------------------------
*! (C) Copyright 2000, 2002 Axis Communications AB, LUND, SWEDEN
*!***************************************************************************/
/* $Id: cgiparse.c,v 1.6 2002-10-03 18:41:00 johana Exp $ */

/********************** INCLUDE FILES SECTION *******************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


#include "cgiparse.h"

/********************** CONSTANT AND MACRO SECTION **************************/

#define D(x)

/********************** TYPE DEFINITION SECTION *****************************/

/********************** LOCAL FUNCTION DECLARATION SECTION ******************/

static unsigned int url_param_fix(unsigned char *s);


/********************** GLOBAL VARIABLE DECLARATION SECTION *****************/

/********************** LOCAL VARIABLE DECLARATION SECTION ******************/

/********************** FUNCTION DEFINITION SECTION *************************/

/*#***************************************************************************
*# FUNCTION NAME: url_param_fix
*#
*# PARAMETERS   : unsigned char* s
*#
*# RETURNS      : true size of the parameter
*#
*# SIDE EFFECTS :
*#
*# DESCRIPTION  : Parses s:
*#                The parameters are separated by '&':s, '+' should be
*#                replaced by ' ', changes "%xx" to '\xx' (hex coded) and
*#                packs the following chars
*#
*#----------------------------------------------------------------------------
*# HISTORY
*#
*# DATE         NAME               CHANGES
*# ----         ----               -------
*# Mar  ? 1997  Johan Adolfsson    Initial version
*# Sep 16 1997  Fredrik Hugosson   Changed int's to uword's and made static.
*#
*#***************************************************************************/
static unsigned int url_param_fix(unsigned char *s)
{
  unsigned int size = 0;
  unsigned int i = 0;

  while (s[i] != '\0')
  {
    if (s[i] == '%')
    {
      i++;
      s[size] = hexbyte(&s[i]);
      i++;
    }
    else
    {
      s[size] = s[i];
    }
    i++;
    size++;
  }
  if (size > 0)
  {
    s[size] = '\0';
  }

  return size;
} /* url_param_fix */

/*#***************************************************************************
*# FUNCTION NAME: url_parameter_type_init
*#
*# PARAMETERS   : const unsigned char* raw_parameter
*#
*# RETURNS      : struct url_parameter_type* or NULL of out of memory
*#
*# SIDE EFFECTS :
*#
*# DESCRIPTION  : Parses raw_parameter and creates a struct url_parameter_type
*#
*#----------------------------------------------------------------------------
*# HISTORY
*#
*# DATE         NAME               CHANGES
*# ----         ----               -------
*# Feb 12 2000  Johan Adolfsson    Initial version
*#
*#***************************************************************************/
struct url_parameter_type* url_parameter_type_init(const char *raw_parameter)
{
  unsigned char *cpy;
  cpy = (unsigned char*)malloc(strlen(raw_parameter) + 1);
  D(printf("init: raw: '%s'\n", raw_parameter));  
  if (cpy == NULL)
  {
    return NULL;
  }
  strcpy(cpy, raw_parameter);
  return  url_parameter_type_init_no_copy(cpy);
}


/*#***************************************************************************
*# FUNCTION NAME: url_parameter_type_init_no_copy
*#
*# PARAMETERS   : unsigned char* raw_parameter
*#
*# RETURNS      : struct url_parameter_type* or NULL of out of memory
*#
*# SIDE EFFECTS :
*#
*# DESCRIPTION  : Parses raw_parameter and creates a struct url_parameter_type
*#
*#----------------------------------------------------------------------------
*# HISTORY
*#
*# DATE         NAME               CHANGES
*# ----         ----               -------
*# Mar  ? 1997  Johan Adolfsson    Initial version
*# Apr  1 1997  Johan Adolfsson    Made url_parameter_type_init handle strange
*#                                 input better.
*# Sep 16 1997  Fredrik Hugosson   Changed int's to uword's.
*# Aug 20 1999  Markus Wejrot      ax_malloc -> malloc
*# Oct  3 2002  Johan Adolfsson    Handle raw_header = "" nice
*#
*#***************************************************************************/
struct url_parameter_type* url_parameter_type_init_no_copy(char *raw_parameter)
{
  struct url_parameter_type* up_p;
  static const char data_not_set[] = "";
  unsigned int i;
  assert(raw_parameter);

  D(printf("init_no_copy: raw: '%s'\n", raw_parameter));

  up_p = (struct url_parameter_type*)
    malloc(sizeof(struct url_parameter_type));
  if (up_p != NULL)
  {
    up_p->data = raw_parameter;
    
    D(printf("data: %i %s\n", strlen(raw_parameter), raw_parameter));
    if (up_p->data != NULL)
    {
   /*   printf("up_p->data:'%s' \n", up_p->data); */
      /* par1=p1&par2=p2 */
      /* count the number of '&' */
      up_p->nbr_of_pairs = 1;
      
      i = 0;

      while(up_p->data[i] != '\0')
      {
        if (up_p->data[i] == '&')
        {
          up_p->nbr_of_pairs++;
        }
        i++;
      }
      D(printf("pairs: %i\n", up_p->nbr_of_pairs));
      up_p->pairs = (struct url_parameter_pair_type*)
        malloc(up_p->nbr_of_pairs * sizeof(struct url_parameter_pair_type));
      
      if (up_p->pairs != NULL)
      {
        unsigned int pair_nbr = 0;

        /* Change '+' to ' '  and set pointers in pairs */
        i = 0;
        up_p->pairs[pair_nbr].key = &up_p->data[i];
        up_p->pairs[pair_nbr].data = data_not_set;

        while (up_p->data[i] != '\0')
        {
          switch (up_p->data[i])
          {
            case '+':
              up_p->data[i] = ' ';
              i++;
              break;

           case '=':
              up_p->data[i] = '\0';
              i++;
              up_p->pairs[pair_nbr].data = &up_p->data[i];
              while ((up_p->data[i] != '\0') && (up_p->data[i] != '&'))
              {
                if (up_p->data[i] == '+')
                {
                  up_p->data[i] = ' ';
                }
                i++;
              }
              break;

           case '&':
             up_p->data[i] = '\0';
             i++;
             pair_nbr++;
             if (pair_nbr < up_p->nbr_of_pairs)
             {
               up_p->pairs[pair_nbr].key = &up_p->data[i];
               up_p->pairs[pair_nbr].data = data_not_set;
               while ((up_p->data[i] != '\0') &&
                      (up_p->data[i] != '=') &&
                      (up_p->data[i] != '&'))
               {
                 if (up_p->data[i] == '+')
                 {
                   up_p->data[i] = ' ';
                 }
                 i++;
               }
             }
             break;

           default:
             i++;
             break;

          } /* switch */

        }
        for (pair_nbr = 0; pair_nbr < up_p->nbr_of_pairs; pair_nbr++)
        {
          (void)url_param_fix((char*)up_p->pairs[pair_nbr].key);
          up_p->pairs[pair_nbr].data_size =
            url_param_fix((char*)up_p->pairs[pair_nbr].data);
        }
      }
      else
      {
        /* failed to allocate */
        free(up_p->data);
        free(up_p);
        up_p = NULL;
      }
    }
  }
  return up_p;
} /* url_parameter_type_init */

/*#***************************************************************************
*# FUNCTION NAME: url_parameter_type_init_header
*#
*# PARAMETERS   : char* raw_header - must be dynamically allocated.
*#
*# RETURNS      : struct url_parameter_type* or NULL of out of memory
*#
*# SIDE EFFECTS : Will modify raw_header and free it when done.
*#
*# DESCRIPTION  : Parses raw_header and creates a struct url_parameter_type
*#                raw_header is a complete HTTP header
*#
*#----------------------------------------------------------------------------
*# HISTORY
*#
*# DATE         NAME               CHANGES
*# ----         ----               -------
*# Mar  3 1999  Johan Adolfsson    Initial version
*#
*#***************************************************************************/
struct url_parameter_type* url_parameter_type_init_header(char* raw_header)
{
  /* Header is of form:
  * GET url HTTP/x.x
  * Content-Type: xx
  * Content-Length: 12345
  * Accept: * / *
  *
  *
  */
  struct url_parameter_type* up_p;
  unsigned int i;
  assert(raw_header);

  up_p = (struct url_parameter_type*)
    malloc(sizeof(struct url_parameter_type));
  if (up_p != NULL)
  {
    up_p->data = raw_header;
    /* count the number of '&' */
    up_p->nbr_of_pairs = 1;
    i = 0;

    while(up_p->data[i] != '\0')
    {
      if (up_p->data[i] == '\n')
      {
        up_p->nbr_of_pairs++;
      }
      i++;
    }

    up_p->pairs = (struct url_parameter_pair_type*)
      malloc(up_p->nbr_of_pairs * sizeof(struct url_parameter_pair_type));

    if (up_p->pairs != NULL)
    {
      unsigned int pair_nbr = 0;
      unsigned char* data = up_p->data;
      unsigned char* header_sep;
      /* First separate each row and put the key pointer to the start of */
      /* replace the CR and LF's with '\0' */
      for (pair_nbr = 0; pair_nbr < up_p->nbr_of_pairs; pair_nbr++)
      {
        up_p->pairs[pair_nbr].key = data;
        /* Find next '\n' */
        header_sep = strchr(data, '\n');
        if (header_sep != NULL)
        {
          *header_sep = '\0';
          data = header_sep + 1;
          header_sep--;
          if (*header_sep == '\r')
          {
            *header_sep = '\0';
          }
        }
      }
      /* Now split each header separated by ':' and put the data pointer to
      ** first non space after the ':'
      */
      for (pair_nbr = 0; pair_nbr < up_p->nbr_of_pairs; pair_nbr++)
      {
        data = (char*)up_p->pairs[pair_nbr].key;
        /* Find next ':' */
        header_sep = strchr(data, ':');
        if (header_sep != NULL)
        {
          *header_sep = '\0'; /* Terminate key */
          data = header_sep + 1;
          while (*data == ' ')
          {
            data++;
          }
          up_p->pairs[pair_nbr].data = data;
          up_p->pairs[pair_nbr].data_size = strlen(data);
        }
        else
        {
          /* No colon found, use space */
          header_sep = strchr(data, ' ');
          if (header_sep != NULL)
          {
            *header_sep = '\0'; /* Terminate key */
            data = header_sep + 1;
            while (*data == ' ')
            {
              data++;
            }
            up_p->pairs[pair_nbr].data = data;
            up_p->pairs[pair_nbr].data_size = strlen(data);
          }
          else
          {
            /* No space, let key and data point to same space */
            up_p->pairs[pair_nbr].data = data;
            up_p->pairs[pair_nbr].data_size = strlen(data);
          }
        }
      }
    }
    else
    {
      /* failed to allocate */
      free(up_p->data);
      free(up_p);
      up_p = NULL;
    }
  }
  return up_p;
} /* url_parameter_type_init */


/*#***************************************************************************
*# FUNCTION NAME: url_parameter_type_done
*#
*# PARAMETERS   : struct url_parameter_type* up_p
*#
*# RETURNS      :
*#
*# SIDE EFFECTS :
*#
*# DESCRIPTION  : Free the memory allocated by up_p
*#
*#----------------------------------------------------------------------------
*# HISTORY
*#
*# DATE         NAME               CHANGES
*# ----         ----               -------
*# Mar  ? 1997  Johan Adolfsson    Initial version
*# Apr  1 1997  Johan Adolfsson    Created this historylist
*#
*#***************************************************************************/
void url_parameter_type_done(struct url_parameter_type* up_p)
{
  assert(up_p);
  if (up_p != NULL)
  {
    if (up_p->data != NULL)
    {
      free(up_p->data);
      up_p->data = NULL;
    }
    if (up_p->pairs != NULL)
    {
      free(up_p->pairs);
      up_p->pairs = NULL;
    }
    free(up_p);
  }
} /* url_parameter_type_done */


/*#***************************************************************************
*# FUNCTION NAME: url_parameter_get_data
*#
*# PARAMETERS   : struct url_parameter_type* up_p
*#                char* key
*#                unsigned int *size_p
*# RETURNS      : unsigned char*, pointer to data field, or NULL
*#                updates *size_p
*#
*# SIDE EFFECTS :
*#
*# DESCRIPTION  : Returns pointer to the data field for the key,
*#                NULL if not found
*#                if size_p!=NULL it updates the *size_p with the length of
*#                data_field
*#
*#----------------------------------------------------------------------------
*# HISTORY
*#
*# DATE         NAME               CHANGES
*# ----         ----               -------
*# Mar  ? 1997  Johan Adolfsson    Initial version
*# Apr  1 1997  Johan Adolfsson    Created this historylist
*# Sep 16 1997  Fredrik Hugosson   Changed int's to uword's.
*# Mar  3 2000  Johan Adolfsson    Back to ints again
*#
*#***************************************************************************/
const unsigned char * 
url_parameter_get_data(const struct url_parameter_type *up_p,
                       const char *key,
                       unsigned long int *size_p)
{
  unsigned int pair_nbr=0;
  assert(up_p);
  assert(key);

  while ((pair_nbr < up_p->nbr_of_pairs) &&
         (strcmp(up_p->pairs[pair_nbr].key,key) != 0))
  {
    pair_nbr++;
  }

  if (pair_nbr < up_p->nbr_of_pairs)
  {
    if (size_p != NULL)
    {
      *size_p = up_p->pairs[pair_nbr].data_size;
    }
    return up_p->pairs[pair_nbr].data;
  }
  return NULL;
} /* url_parameter_get_data */



/*#***************************************************************************
*# FUNCTION NAME: hexbyte
*#
*# PARAMETERS   : const char* s - String to get hex byte from
*#
*# RETURNS      : Byte value of the first two chars in s
*#
*# SIDE EFFECTS : None
*#
*# DESCRIPTION  : Returns the value of the two first (hexadecimal) chars in s.
*#                Non hex chars will give result 0 for that digit.
*#----------------------------------------------------------------------------
*# HISTORY
*#
*# DATE         NAME               CHANGES
*# ----         ----               -------
*# Mar  3 1999  Johan Adolfsson    Initial version
*#
*#***************************************************************************/
unsigned char hexbyte(const char *s)
{
  assert(s);
  {
    char c1 = toupper(s[0]);
    char c2 = toupper(s[1]);
    unsigned char sum = 0;

    if (isdigit(c1))
    {
      sum = c1 - '0';
    }
    else
    {
      if (('A' <= c1) && (c1 <= 'F'))
      {
        sum = 10 + c1 - 'A';
      }
    }
    /* Shift up Most Significant hexdigit */
    sum = sum << 4;

    if (isdigit(c2))
    {
      sum += c2 - '0';
    }
    else
    {
      if (('A' <= c2) && (c2 <= 'F'))
      {
        sum += 10 + c2 - 'A';
      }
    }
    return sum;
  }
} /* hexbyte */


/********************** END OF FILE cgiparse.c ******************************/
