/*!***************************************************************************
*!
*! FILE NAME  : cgiparse.h (based on http_pkt.h)
*!
*! DESCRIPTION: Parse and CGI queries
*!
*! FUNCTIONS  : url_parameter_type_init
*! (EXPORTED)   url_parameter_type_init_header
*!              url_parameter_type_done
*!              url_parameter_get_data
*!
*! ---------------------------------------------------------------------------
*! HISTORY
*!
*! DATE         NAME               CHANGES
*! ----         ----               -------
*! Feb 12 2000  Johan Adolfsson    Initial version
*!
*! ---------------------------------------------------------------------------
*! (C) Copyright 2000, Axis Communications AB, LUND, SWEDEN
*!***************************************************************************/
/* $Id: cgiparse.h,v 1.2 2002-10-03 14:15:06 johana Exp $ */

#ifndef CGIPARSE_H
#define CGIPARSE_H

/********************** INCLUDE FILES SECTION *******************************/


/********************** CONSTANT AND MACRO SECTION **************************/

/********************** TYPE DEFINITION SECTION *****************************/


/* types for handling parameters specified in a HTTP GET:
e.g. "/serial?data=Hello" */
typedef struct url_parameter_pair_type
{
  const char *key;  /* pointer to key */
  const unsigned char *data; /* pointer to data  */
  unsigned long int data_size; /* length of data (may contain '\0')*/
}url_parameter_pair_type;

typedef struct url_parameter_type
{
  unsigned char *data; /* pointer to allocated data  */
  unsigned int nbr_of_pairs; /* number of pairs in array */
  struct url_parameter_pair_type *pairs; /* pointer to array */
}url_parameter_type;



/********************** EXPORTED FUNCTION DECLARATION SECTION ***************/



struct url_parameter_type*
url_parameter_type_init(const char *raw_parameter);
/* Parses raw_parameter in HTTP url format and
** creates a struct url_parameter_type, return NULL if out of mem
*/

struct url_parameter_type*
url_parameter_type_init_no_copy(char *raw_parameter);
/* Parses raw_parameter in HTTP url format and
** creates a struct url_parameter_type, return NULL if out of mem
*/

struct url_parameter_type*
url_parameter_type_init_header( char *raw_header);
/* Parses raw_header in HTTP headerand creates a struct url_parameter_type
 * Will modify raw_header during parsing and will free it when _done is called
 * raw_header must be dynamically allocated.
 */

void url_parameter_type_done(struct url_parameter_type *up_p);
/* Free the memory allocated by up_p and up_p it self */

const unsigned char * 
url_parameter_get_data(const struct url_parameter_type *up_p,
                       const char  *key,
                       unsigned long int *size_p);
                                   
unsigned char hexbyte(const char *s);


#endif /* CGIPARSE_H */
/********************** END OF FILE cgiparse.h ******************************/

