/*!***************************************************************************
*!
*! FILE NAME  : editcgi.c
*!
*! DESCRIPTION: Simple CGI based "Editor" and file browser
*!              by Johan Adolfsson <Johan.Adolfsson@axis.com>
*!              Edits files (should anly be allowed to run as root)
*!              editfile.cgi?file=path
*! 
*!
*!
*! FUNCTIONS  : main 
*! (EXTERNAL)   write_to_fd
*!
*! FUNCTIONS  : 
*! (LOCAL)
*!
*! ---------------------------------------------------------------------------
*! HISTORY
*!
*! DATE         NAME               CHANGES
*! ----         ----               -------
*! Feb 12 2000  Johan Adolfsson    Initial version
*!
*! This file is distributed under the Gnu Public License (GPL),
*! please see the file LICENSE for further information.
*! 
*! ---------------------------------------------------------------------------
*! (C) Copyright 2000, Axis Communications AB, LUND, SWEDEN
*!***************************************************************************/
/* $Id: editcgi.c,v 1.10 2003-12-03 08:29:07 johana Exp $ */

/****************** INCLUDE FILES SECTION ***********************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>

#include "ls.h"
#include "cgiparse.h"

/****************** CONSTANT AND MACRO SECTION ******************************/

#define DPOST(x) 
#define DFILE(x) x
#define BUF_SIZE 1024

/****************** TYPE DEFINITION SECTION *********************************/

/****************** LOCAL FUNCTION DECLARATION SECTION **********************/

/****************** GLOBAL VARIABLE DECLARATION SECTION *********************/
extern char **environ;

/****************** LOCAL VARIABLE DECLARATION SECTION **********************/
static int debug = 0;               /* Print debug info if 1 */
static int show_env = 0;            /* Show environment variables if 1 */
static unsigned char glob_buf[BUF_SIZE];

/****************** FUNCTION DEFINITION SECTION *****************************/

/* This is used by ls */
int write_to_fd(int fd, const char *buf, int len)
{
  printf("%.*s", len, buf);
  return len;
}



int main(int argc, char **argv)
{
  int i;
  const char *method;
  const char *query;
  const char *uri;
  const char *script_name;
  struct url_parameter_type *up_p;
  
  method = getenv("REQUEST_METHOD");
  query = getenv("QUERY_STRING");
  script_name = getenv("SCRIPT_NAME");
  if (query == NULL)
  {
    query = "";
    if (argc > 1)
    {
      query = argv[1];
    }
  }
  if (script_name == NULL)
  {
    script_name = argv[0];
  }
  
  /* Print mandatory Content-Type */
  printf("Content-Type: text/html\n\n");

  up_p = url_parameter_type_init(query);

  if (url_parameter_get_data(up_p, "debug", NULL))
  {
    debug = 1;
  }
  if (url_parameter_get_data(up_p, "env", NULL))
  {
    show_env = 1;
  }
  
  
  uri = getenv("REQUEST_URI");

  if (method == NULL)
  {
    /* Not a CGI! */
    printf("This program should be run as a CGI from a web server, "
           "faking GET!\n");
    query = argv[1];
    method = "GET";
   
  }


  
  /* Print body */
  printf("<html><body>\n");


  if (debug)
  {
    /* Important ENV variables */
    printf("Important ENV:\n");
    printf("method: %s<br>\n", method);
    printf("query: %s<br>\n", query);
    printf("uri: %s<br>\n", uri);
  }
  
  if (show_env)
  {
    printf("<pre>\n");
    printf("ENV:\n");  
    i=0;
    while (environ[i] != NULL)
    {
      printf("%s\n", environ[i]);
      i++;
      
    }
    printf("</pre>\n");
  }
  
  if (strcmp(method, "GET") == 0)
  {
    const unsigned char *fname;
    fname = url_parameter_get_data(up_p, "file", NULL);
    if (fname == NULL)
    {
      printf("You must specify a name of file to edit or press Submit to browse:<br>\n");
      
      printf("<form method=\"GET\">\n"
             "File name: <input type=\"text\" name=\"file\" "
             "size=50 maxlength=255 value=\"\">\n"
             );
      printf("<input type=submit value=\"Submit\">\n"
             "</form><br>\n"
             );
    }
    else
    {
      struct stat file_stat;
      if (strlen(fname) == 0)
      {
        fname = "/"; /* Handle root */
      }
      
      if (stat(fname, &file_stat) != 0)
      {
        /* Some error */
        printf("Failed to get status of %s\n", fname);
      }
      else
      {
        if (S_ISDIR(file_stat.st_mode))
        {
          DIR *dirp;
          printf("Directory: %s\n", fname);
          dirp = opendir(fname);
          if(!dirp) {
            printf("Failed to read dir: %s\n", fname);
          }
          else
          {
            char buf [200];
            sprintf(buf, "%s?file=",script_name);
            
            printf("<pre>\n");
            /* closedir(dirp) is called by ls */
            ls(fname, dirp, STDOUT_FILENO,1, buf);
            printf("</pre>\n");
          }
        }
        else
        {
          int fin;
          ssize_t len;
          
          fin = open(fname, O_RDONLY);
          
          printf("File: %s  Length: %lu bytes\n", fname, file_stat.st_size);
          printf("<a href=\"%s\">[Select new file]</a>\n", script_name);
          
          if (fin < 0)
          {
            printf("Failed to open %s !\n", fname);
          }
          else
          { 
            printf("<form method=\"POST\">\n"
                   "<table><tr><td>"
                   "Save as: "
                   "<input type=text name=\"save_file\" value=\"%s\" "
                   "size=30 maxlength=255></td>\n"
                   "<td>Mode: <input type=text name=\"mode\" value=\"0%lo\">"
                   "</td>\n"
                   "<td>Convert CRLF to LF: <input type=checkbox "
                   "name=\"convert_crlf_to_lf\" value=on checked>"
                   "</td>\n</tr>\n"
                   "<tr><td colspan=3 align=center>"
                   "<input type=submit value=\" Save file \">"
                   "</td>"
                   "</tr>\n"
                   "<tr><td colspan=3>"
                   "<textarea rows=20 cols=80 name=content>\n",
                   fname, 
                   (unsigned long)file_stat.st_mode
                   );
            len = read(fin, glob_buf, BUF_SIZE);
            
            while(len > 0)
            {
              printf("%.*s", (int)len, glob_buf); /* printf work's but is it safe? */
              /*write(STDOUT_FILENO, buf, len); */ /* This won't work */
              len = read(fin, glob_buf, BUF_SIZE);
            }
            close(fin);
            printf("</textarea>\n"
                   "</td></tr><tr><td colspan=3 align=center>"
                   "<input type=submit value=\" Save file \">\n"
                   "</td></tr></table>"
                   "</form><br>\n");
          }
        }
      }
    }
  }
  else if (strcmp(method, "POST") == 0)
  {
    unsigned long content_length = 0;
    struct url_parameter_type *body_p;
    ssize_t len = 0;
    ssize_t tmp_len = 0;    
    const char *s;
    const char *save_file;
    const char *file_mode;
    const char *content;
    unsigned long int file_content_len = 0;
    int convert_crlf_to_lf = 0;
    int fout;
    char* body;
    
    DPOST(printf("POST<br>\n"));
    DPOST(printf("query: %s<br>\n", query));

    s = getenv("CONTENT_LENGTH");
    content_length= atol(s);
    DPOST(printf("Content-Length: %lu\n", content_length));
    body = malloc(content_length + 1);
    while (len < content_length)
    {
      DPOST(printf("Read body\n"));
      tmp_len = read(fileno(stdin), &body[len], content_length-len);
      if (tmp_len == -1)
      {
        perror("read stdin:");
      }
      
      len += tmp_len;
      DPOST(printf("len: %i tmp_len: %i\n", len, tmp_len));
    }
    body[content_length] = '\0';
    DPOST(printf("BODY:\n<pre>%s</pre>\n", body));
    body_p = url_parameter_type_init_no_copy(body);
    save_file = url_parameter_get_data(body_p, "save_file", NULL);
    file_mode = url_parameter_get_data(body_p, "mode", NULL);
    content = url_parameter_get_data(body_p, "content", &file_content_len);
    if (url_parameter_get_data(body_p, "convert_crlf_to_lf", NULL) != NULL)
    {
      convert_crlf_to_lf = 1;
    }
    if (save_file != NULL && content != NULL)
    {
      mode_t mode = strtol(file_mode, NULL, 0);
      umask(0777);

      printf("File: %s  Length: %lu bytes  Mode 0%o\n", save_file, file_content_len, (unsigned int)mode);
      printf("<pre>\n");
      fout = open(save_file, O_WRONLY | O_CREAT | O_TRUNC, mode);
      if (fout > 0)
      {
        int ret;
        int err = 0;
        unsigned int tot = 0;
        while (err == 0 && tot < file_content_len)
        {
          if (convert_crlf_to_lf)
          {
            char *lf_pos = strchr(&content[tot], '\n');
            if (lf_pos != NULL && 
                lf_pos != &content[tot] && 
                *(lf_pos-1) == '\r')
            {
              int size = lf_pos - &content[tot] - 1;
              
              DPOST(printf("Found CRLF!\n"));
              if (size)
              {
                ret = write(fout, &content[tot], size);
                if (ret < 0)
                {
                  err++;
                }
              }
              
              ret = write(fout, "\n", 1);
              if (ret < 0)
              {
                err++;
              }
              DPOST(printf("%.*s\n", size, &content[tot])); /* Write \n */
              tot += size + 2; /* before CRLF + the CRLF */
            }
            else
            {
              ret = write(fout, &content[tot], file_content_len - tot);
              if (ret < 0)
              {
                err++;
              }
              DPOST(printf("%.*s", file_content_len - tot, &content[tot]));
              tot += ret;
            }
          }
          else
          {
            ret = write(fout, &content[tot], file_content_len);
            if (ret < 0)
            {
              err++;
            }
            DPOST(printf("%.*s", file_content_len, &content[tot]));
            tot += ret;
          }
        } /* while */
        printf("Wrote %lu bytes\n", (unsigned long)tot);
        fsync(fout);
        close(fout);
        if (err > 0)
        {
          printf("Error: %i detected during write!\n", errno);
        }
        /* We need to do chmod if the file already existed */
        chmod(save_file, mode);
      } /* if (fout > 0) */
      else
      {
        printf("Failed to open/create file! (%i) %s <br>\n", 
               errno, strerror(errno));
      }
      printf("</pre>\n");
    }
    else
    {
      /* Something went wrong.. */
      printf("Something went wrong  (%i) - can't save file!\n<br>", errno);
    }
    {
      char *name = glob_buf;
      char *last_slash;
      
      strcpy(name, save_file);
      last_slash = strrchr(name, '/');
      if (last_slash != NULL)
      {
        /* Remove filename - just path left */ 
        last_slash[1] = '\0';
      }
      
      printf("<a href=\"%s\">[Select new file]</a> <br>\n", script_name);
      printf("<a href=\"%s?file=%s\">%s</a> <br>\n", script_name, name, name);
    }
    
    url_parameter_type_done(body_p);
  }
  else
  {
    printf("Unhandled method: %s<br>\n", method);
  }
  
  printf("</body></html>\n");
  url_parameter_type_done(up_p);
  
  return 0;
}

/****************** END OF FILE editcgi.c **********************************/
