/* $Id: ls.c,v 1.4 2003-01-28 21:20:45 johana Exp $
 * This is taken from sftpd - but with HTML link functionality added.
 * simple "ls -l" functionality. parts borrowed from here and there.
 *
 */

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef	S_ISLNK
#define	LSTAT	lstat
#else
#define	LSTAT	stat
#endif

#define PATHLEN 512

extern int write_to_fd(int fd, const char *buf, int len);

static int lsfile(char *fullname, const char *name,
                  struct stat *statbuf, int fd, int do_HTML_links, 
                  const char *htmllink_prefix);

static char *
modestring(int mode)
{
        static  char    buf[12];

        strcpy(buf, "----------");

        /*
         * Fill in the file type.
         */
        if (S_ISDIR(mode))
                buf[0] = 'd';
        if (S_ISCHR(mode))
                buf[0] = 'c';
        if (S_ISBLK(mode))
                buf[0] = 'b';
        if (S_ISFIFO(mode))
                buf[0] = 'p';
#ifdef  S_ISLNK
        if (S_ISLNK(mode))
                buf[0] = 'l';
#endif
#ifdef  S_ISSOCK
        if (S_ISSOCK(mode))
                buf[0] = 's';
#endif

        /*
         * Now fill in the normal file permissions.
         */
        if (mode & S_IRUSR)
                buf[1] = 'r';
        if (mode & S_IWUSR)
                buf[2] = 'w';
        if (mode & S_IXUSR)
                buf[3] = 'x';
        if (mode & S_IRGRP)
                buf[4] = 'r';
        if (mode & S_IWGRP)
                buf[5] = 'w';
        if (mode & S_IXGRP)
                buf[6] = 'x';
        if (mode & S_IROTH)
                buf[7] = 'r';
        if (mode & S_IWOTH)
                buf[8] = 'w';
        if (mode & S_IXOTH)
                buf[9] = 'x';

        /*
         * Finally fill in magic stuff like suid and sticky text.
         */
        if (mode & S_ISUID)
                buf[3] = ((mode & S_IXUSR) ? 's' : 'S');
        if (mode & S_ISGID)
                buf[6] = ((mode & S_IXGRP) ? 's' : 'S');
        if (mode & S_ISVTX)
                buf[9] = ((mode & S_IXOTH) ? 't' : 'T');

        return buf;
}
struct stat updir_statbuf = { 0 };


int
ls(const char *dirname, DIR *dirp, int fd, int do_HTML_links,    
   const char *htmllink_prefix)
{
	int		endslash;
	struct	dirent	*dp;
	char            fullname[PATHLEN];
	struct	stat	statbuf;
        int first = 1;
	
	endslash = (*dirname && (dirname[strlen(dirname) - 1] == '/'));

	while ((dp = readdir(dirp)) != NULL) {
	start_again:
		strncpy(fullname, dirname, PATHLEN - 1);
		fullname[PATHLEN - 1] = '\0';
		if (!endslash)
			strcat(fullname, "/");

		/* cramfs don't have '.' and '..' entries so lets
		   fake them */
		if (first && 
		    strcmp(".", dp->d_name) != 0 &&
		    strcmp("..", dp->d_name) != 0) {
			/* Lets add fake '.' and '..' links */
			strcat(fullname,".");
			lsfile(fullname, ".", &updir_statbuf, fd, 
			       do_HTML_links, htmllink_prefix);
			strcat(fullname,".");
			lsfile(fullname, "..", &updir_statbuf, fd, 
			       do_HTML_links, htmllink_prefix);
			
			first = 0;
			goto start_again;
		} else {
			first = 0;
		}
		
		
		strcat(fullname, dp->d_name);

		if (LSTAT(fullname, &statbuf) < 0) {
			closedir(dirp);
			perror(fullname);
			return -1;
		}

		if(lsfile(fullname, dp->d_name, &statbuf, fd, do_HTML_links,
			  htmllink_prefix) < 0) {
			closedir(dirp);
			return -1;
		}
	}

	closedir(dirp);
	
	return 0;
}

/*
 * Get the time to be used for a file.
 * This is down to the minute for new files, but only the date for old files.
 * The string is returned from a static buffer, and so is overwritten for
 * each call.
 */

static char *
timestring(long t)
{
        long            now;
        char            *str;
        static  char    buf[26];

        time(&now);

        str = ctime(&t);

        strcpy(buf, &str[4]);
        buf[12] = '\0';

        if ((t > now) || (t < now - 365*24*60*60L)) {
                strcpy(&buf[7], &str[20]);
                buf[11] = '\0';
        }

        return buf;
}

static int
lsfile(char *fullname, 
       const char *name, 
       struct stat *statbuf, 
       int fd, 
       int do_HTML_links, 
       const char *htmllink_prefix)
{
	char		*cp;
	struct	passwd	*pwd;
	struct	group	*grp;
	static char     buf[PATHLEN];
	static	char	username[12];
	static	int	userid;
	static	int	useridknown = 0;
	static	char	groupname[12];
	static	int	groupid;
	static	int	groupidknown = 0;

	memset(buf,0,PATHLEN);
	cp = buf;
	*cp = '\0';

	strcpy(cp, modestring(statbuf->st_mode));
	cp += strlen(cp);

	sprintf(cp, "%3d ", statbuf->st_nlink);
	cp += strlen(cp);

	/* optimize a bit and cache 1 lookup */

	if (!useridknown || (statbuf->st_uid != userid)) {
		pwd = getpwuid(statbuf->st_uid);
		if (pwd)
			strcpy(username, pwd->pw_name);
		else
			sprintf(username, "%d", statbuf->st_uid);
		userid = statbuf->st_uid;
		useridknown++;
	}

	sprintf(cp, "%-8s ", username);
	cp += strlen(cp);

	if (!groupidknown || (statbuf->st_gid != groupid)) {
		grp = getgrgid(statbuf->st_gid);
		if (grp)
			strcpy(groupname, grp->gr_name);
		else
			sprintf(groupname, "%d", statbuf->st_gid);
		groupid = statbuf->st_gid;
		groupidknown++;
	}

	sprintf(cp, "%-8s ", groupname);
	cp += strlen(cp);

	if (S_ISBLK(statbuf->st_mode) || S_ISCHR(statbuf->st_mode))
		sprintf(cp, "%3d, %3d ", (int)(statbuf->st_rdev >> 8),
			(int)(statbuf->st_rdev & 0xff));
	else
		sprintf(cp, "%8d ", (int)statbuf->st_size);
	cp += strlen(cp);
	
	sprintf(cp, " %-12s ", timestring(statbuf->st_mtime));

	if(write_to_fd(fd, buf, strlen(buf)) < 0)
		return -1;
  if (do_HTML_links)
  {
    const char *linkname = fullname;
    int len = strlen(fullname);
    
    if (len == 0)
    {
      linkname = "/"; /* Handle root */
    }
    else if (strcmp(".", name) == 0) {
	    /* Remove the . in fullname */
	    /* fullname is "path/dir/." so first remove last /.*/
	    fullname[len-1] = '\0';
    } 
    else if (strcmp("..", name) == 0) {
	    /* Remove last directory entry in fullname */
	    char *end;
	    /* fullname is "path/dir/.." so first remove last /..*/
	    fullname[len-3] = '\0';
	    end = strrchr(fullname, '/');
	    if (end) {
		    end[1] = '\0';
	    }
    }
    
    sprintf(buf, "<a href=\"%s%s\">", htmllink_prefix, linkname);
    if(write_to_fd(fd, buf, strlen(buf)) < 0)
      return -1;
  }
    
	if(write_to_fd(fd, name, strlen(name)) < 0)
		return -1;
  if (do_HTML_links)
  {
    sprintf(buf, "</a>");
    if(write_to_fd(fd, buf, strlen(buf)) < 0)
      return -1;
  }	

#ifdef	S_ISLNK
	if (S_ISLNK(statbuf->st_mode)) {
		int len;
		char linkname[PATHLEN];
		
		memset(linkname,0,PATHLEN);

		if(write_to_fd(fd, " -> ", 4) < 0)
			return -1;

		len = readlink(fullname, linkname, PATHLEN);

		if (len >= 0) {
			struct	stat	statlinkbuf;

			if(write_to_fd(fd, linkname, len) < 0)
				return -1;

			if (!(LSTAT(linkname, &statlinkbuf) < 0)) {
				if (S_ISDIR(statlinkbuf.st_mode)) {			
					/* add slash only if it's missing */
					if (linkname[len-1] != '/') {
						if(write_to_fd(fd, "/", 1) < 0)
							return -1;
					}
				}
			}
		}
	}
#endif

	return write_to_fd(fd, "\r\n", 2);
}
